use std::path::Path;

use crate::compare::diff::compare_document;
use crate::compare::diff::DiffResult;
use crate::compare::parse::emacs_parse_anonymous_org_document;
use crate::compare::parse::emacs_parse_file_org_document;
use crate::compare::parse::get_emacs_version;
use crate::compare::parse::get_org_mode_version;
use crate::compare::sexp::sexp;
use crate::context::GlobalSettings;
use crate::context::LocalFileAccessInterface;
use crate::parser::parse_file_with_settings;
use crate::parser::parse_with_settings;

pub async fn run_anonymous_compare<P: AsRef<str>>(
    org_contents: P,
) -> Result<bool, Box<dyn std::error::Error>> {
    run_anonymous_compare_with_settings(org_contents, &GlobalSettings::default(), false).await
}

pub async fn run_compare_on_file<P: AsRef<Path>>(
    org_path: P,
) -> Result<bool, Box<dyn std::error::Error>> {
    run_compare_on_file_with_settings(org_path, &GlobalSettings::default(), false).await
}

pub async fn silent_anonymous_compare<P: AsRef<str>>(
    org_contents: P,
) -> Result<bool, Box<dyn std::error::Error>> {
    run_anonymous_compare_with_settings(org_contents, &GlobalSettings::default(), true).await
}

pub async fn silent_compare_on_file<P: AsRef<Path>>(
    org_path: P,
) -> Result<bool, Box<dyn std::error::Error>> {
    run_compare_on_file_with_settings(org_path, &GlobalSettings::default(), true).await
}

pub async fn run_anonymous_compare_with_settings<'g, 's, P: AsRef<str>>(
    org_contents: P,
    global_settings: &GlobalSettings<'g, 's>,
    silent: bool,
) -> Result<bool, Box<dyn std::error::Error>> {
    // TODO: This is a work-around to pretend that dos line endings do not exist. It would be better to handle the difference in line endings.
    let org_contents = org_contents.as_ref().replace("\r\n", "\n");
    let org_contents = org_contents.as_str();
    if !silent {
        print_versions().await?;
    }
    let rust_parsed = parse_with_settings(org_contents, global_settings)?;
    let org_sexp = emacs_parse_anonymous_org_document(org_contents, global_settings).await?;
    let (_remaining, parsed_sexp) = sexp(org_sexp.as_str()).map_err(|e| e.to_string())?;

    if !silent {
        println!("{}\n\n\n", org_contents);
        println!("{}", org_sexp);
        println!("{:#?}", rust_parsed);
    }

    // We do the diffing after printing out both parsed forms in case the diffing panics
    let diff_result = compare_document(&parsed_sexp, &rust_parsed)?;
    if !silent {
        diff_result.print(org_contents)?;
    }

    if diff_result.is_bad() {
        return Ok(false);
    } else if !silent {
        println!(
            "{color}Entire document passes.{reset}",
            color = DiffResult::foreground_color(0, 255, 0),
            reset = DiffResult::reset_color(),
        );
    }

    Ok(true)
}

pub async fn run_compare_on_file_with_settings<'g, 's, P: AsRef<Path>>(
    org_path: P,
    global_settings: &GlobalSettings<'g, 's>,
    silent: bool,
) -> Result<bool, Box<dyn std::error::Error>> {
    let org_path = org_path.as_ref();
    if !silent {
        print_versions().await?;
    }
    let parent_directory = org_path
        .parent()
        .ok_or("Should be contained inside a directory.")?;
    let org_contents = std::fs::read_to_string(org_path)?;
    // TODO: This is a work-around to pretend that dos line endings do not exist. It would be better to handle the difference in line endings.
    let org_contents = org_contents.replace("\r\n", "\n");
    let org_contents = org_contents.as_str();
    let file_access_interface = LocalFileAccessInterface {
        working_directory: Some(parent_directory.to_path_buf()),
    };
    let global_settings = {
        let mut global_settings = global_settings.clone();
        global_settings.file_access = &file_access_interface;
        global_settings
    };
    let rust_parsed = parse_file_with_settings(org_contents, &global_settings, Some(org_path))?;
    let org_sexp = emacs_parse_file_org_document(org_path, &global_settings).await?;
    let (_remaining, parsed_sexp) = sexp(org_sexp.as_str()).map_err(|e| e.to_string())?;

    if !silent {
        println!("{}\n\n\n", org_contents);
        println!("{}", org_sexp);
        println!("{:#?}", rust_parsed);
    }

    // We do the diffing after printing out both parsed forms in case the diffing panics
    let diff_result = compare_document(&parsed_sexp, &rust_parsed)?;
    if !silent {
        diff_result.print(org_contents)?;
    }

    if diff_result.is_bad() {
        return Ok(false);
    } else if !silent {
        println!(
            "{color}Entire document passes.{reset}",
            color = DiffResult::foreground_color(0, 255, 0),
            reset = DiffResult::reset_color(),
        );
    }

    Ok(true)
}

async fn print_versions() -> Result<(), Box<dyn std::error::Error>> {
    eprintln!("Using emacs version: {}", get_emacs_version().await?.trim());
    eprintln!(
        "Using org-mode version: {}",
        get_org_mode_version().await?.trim()
    );
    Ok(())
}
